@@ -1,5 +1,5 @@
 /* SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only */
-/* Copyright (c) 2023-2024 Brett Sheffield <bacs@librecast.net> */
+/* Copyright (c) 2023-2025 Brett Sheffield <bacs@librecast.net> */
 
 /* sync.h - Librecast data & file syncing API */
 
@@ -13,6 +13,7 @@
 
 #define NET_LOOPBACK 1
 #define SYNC_SEND_THREADS 1
+#define SYNC_MAX_RECVERS 16 /* maximum number of simultaneous receive groups */
 
 enum {
 	/* flags for lc_sync functions */
@@ -31,15 +32,17 @@ typedef uint8_t hash_t[HASHSIZE];
 typedef struct net_tree_s net_tree_t;
 struct net_tree_s {
 	uint64_t	size;	/* Total size, in bytes */
-	struct timespec	mtime;  /* Time of last modification */
-	struct timespec	atime;  /* Time of last access */
-	mode_t		mode;	/* File type and mode */
-	uid_t		uid;	/* User ID of owner */
-	gid_t		gid;	/* Group ID of owner */
+	uint64_t        mtime_s;/* Time of last modification - sec */
+	uint64_t        mtime_n;/*                           - nsec */
+	uint64_t        atime_s;/* Time of last access       - sec */
+	uint64_t        atime_n;/*                           - nsec */
+	uint32_t	uid;	/* User ID of owner */
+	uint32_t	gid;	/* Group ID of owner */
+	uint16_t	mode;	/* File type and mode */
 	uint16_t	namesz;	/* length of filename */
 				/* name is prefixed to tree */
 				/* (symlink target replaces tree hashes) */
-	uint64_t	pad;
+	uint32_t	pad;
 	unsigned char	tree[];	/* tree or directory hashes */
 };
 
@@ -61,6 +64,9 @@ lc_share_t * lc_share(lc_ctx_t *lctx, mdex_t *mdex, unsigned int ifx, lc_stat_t
 		lc_sync_options_t *opt, int flags);
 void lc_unshare(lc_share_t *share);
 
+lc_share_t *lc_share_acquire(lc_share_t *share);
+void lc_share_release(lc_share_t *share);
+
 ssize_t lc_recvdir(lc_ctx_t *lctx, unsigned char *hash, net_tree_t **dir,
 		lc_stat_t *stats, lc_sync_options_t *opt, int flags);
 ssize_t lc_recvtree(lc_ctx_t *lctx, unsigned char *hash, mtree_t *tree,
