@@ -127,7 +127,7 @@ static void freeifacelist(mld_iface_t *i)
 
 static mld_iface_t *mld_get_iface(const mld_t *mld, const unsigned int ifx)
 {
-	for (mld_iface_t *i = mld->iface; i; i = i->next) {
+	for (mld_iface_t *i = aload(&mld->iface); i; i = aload(&i->next)) {
 		if (i->ifx == ifx) return i;
 	}
 	return NULL;
@@ -218,13 +218,15 @@ static int mld_join_iface(mld_t *mld, unsigned int ifx)
 		if (!(ifa->ifa_flags & IFF_MULTICAST)) {
 			continue;
 		}
-		if (!setsockopt(mld->sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, &req, sizeof(req))) {
+		int sock = aload(&mld->sock);
+		if (!setsockopt(sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, &req, sizeof(req))) {
 			DEBUG("MLD listening on interface %s[%u]\n", ifa->ifa_name, ifx);
 			memcpy(&iface->llink, llocal, sizeof(struct sockaddr_in6));
 		}
 		else {
 			ERROR("%s() setsockopt: %s\n", __func__, strerror(errno));
 		}
+		astor(&mld->sock, sock);
 	}
 	freeifaddrs(ifaddr);
 	/* send MLD2 General Queries */
@@ -265,13 +267,13 @@ int mld_add_iface(mld_t *mld, unsigned int ifx, char *ifname)
 		return (errno = ENODEV), -1;
 	iface = calloc(1, sizeof(mld_iface_t));
 	if (!iface) return -1;
-	iface->next = mld->iface;
+	astor(&iface->next, aload(&mld->iface));
 	iface->ifx = ifx;
 	strcpy(iface->ifname, ifname);
 #if OS_NETBSD
 	iface->bpf = mld_bpf_init(ifname);
 #endif
-	mld->iface = iface;
+	astor(&mld->iface, iface);
 	mld_join_iface(mld, ifx);
 	DEBUG("ifx %u added\n", ifx);
 	mld_watch_event_push(mld, ifx, NULL, MLD_EVENT_IFUP);
@@ -774,7 +776,7 @@ static int mld_is_querier(mld_t *mld, unsigned int ifx, struct timespec *ts)
 	if (!iface) return 0;
 	if (ts) now.tv_sec = ts->tv_sec;
 	else if (clock_gettime(CLOCK_REALTIME, &now) == -1) return -1;
-	return (iface->qseen.tv_sec + MLD2_OTHER_QUERIER_TIMEOUT > now.tv_sec) ? 0 : 1;
+	return ((aload(&iface->qseen.tv_sec) + MLD2_OTHER_QUERIER_TIMEOUT) > now.tv_sec) ? 0 : 1;
 }
 
 void mld_query_msg_free(mld_msg_t *msg)
@@ -878,7 +880,7 @@ static void mld_query_send(mld_t *mld, unsigned int ifx, struct in6_addr *saddr)
 	mld_msg_t msg = {0};
 	mld_iface_t *iface;
 	struct timespec now = {0};
-	int sock = mld->sock;
+	int sock = aload(&mld->sock);
 	int opt = 1;
 
 	iface = mld_get_iface(mld, ifx);
@@ -928,6 +930,7 @@ static void mld_query_send(mld_t *mld, unsigned int ifx, struct in6_addr *saddr)
 	for (int i = 0; i < MLD2_ROBUSTNESS; i++) {
 		if (sendmsg(sock, &msg.msgh, 0) == -1) ERROR("sendmsg\n");
 	}
+	astor(&mld->sock, sock);
 err_mld_query_msg_free:
 	mld_query_msg_free(&msg);
 }
@@ -1021,7 +1024,7 @@ void mld_query_handler(mld_t *mld, unsigned int ifx, struct ip6_hdr *ip6h)
 			if_indextoname(ifx, ifname);
 			DEBUG("Querier detected. Entering Non-Querier state on %s(%u)\n", ifname, ifx);
 		}
-		iface->qseen.tv_sec = now.tv_sec;
+		astor(&iface->qseen.tv_sec, now.tv_sec);
 	}
 }
 
@@ -1363,10 +1366,11 @@ mld_t *mld_start(mld_t *mld)
 #endif
 
 	/* start listen thread and wait for it to initialize */
-	mld->sock = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
+	int sock = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
 	if (mld->sock == -1) goto err_mld_stop;
-	if (setsockopt(mld->sock, IPPROTO_IPV6, IPV6_RECVPKTINFO, &opt, sizeof(opt)))
+	if (setsockopt(sock, IPPROTO_IPV6, IPV6_RECVPKTINFO, &opt, sizeof(opt)))
 		goto err_mld_stop;
+	astor(&mld->sock, sock);
 	mld->job[MLD_LISTEN] = job_push_new(mld->q[MLD_LISTEN],
 			&mld_thread_listen, mld, sizeof(*mld), NULL, 0);
 	if (!mld->job[MLD_LISTEN]) goto err_mld_stop;
