@@ -28,6 +28,16 @@
 
 #define HAVE_NETLINK __linux__
 
+#ifndef aload
+# define aload(x) __atomic_load_n((x), __ATOMIC_ACQUIRE)
+#endif
+#ifndef astor
+# define astor(x, y) __atomic_store_n((x),(y), __ATOMIC_RELEASE)
+#endif
+#ifndef aadd
+# define aadd(x, y) __atomic_fetch_add((x),(y), __ATOMIC_ACQ_REL)
+#endif
+
 typedef enum {
 	MLD_NETLINK,
 	MLD_LISTEN,
