@@ -30,6 +30,11 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <poll.h>
+#ifdef HAVE_LINUX_NETLINK_H
+#include <asm/types.h>
+#include <linux/netlink.h>
+#include <linux/rtnetlink.h>
+#endif
 
 #ifdef HAVE_LIBLCRQ
 # define ESI(row, col) chan->ESI[row * nesi + col]
@@ -89,22 +94,35 @@ lc_ctx_t *lc_channel_ctx(lc_channel_t *chan)
 	return chan->ctx;
 }
 
-#ifdef HAVE_LIBLCRQ
 rq_t *lc_channel_rq(lc_channel_t *chan)
 {
+#ifdef HAVE_LIBLCRQ
 	return chan->rq;
+#else
+	(void)chan;
+	return (errno = ENOTSUP), NULL;
+#endif
 }
 
 int lc_ctx_rq_overhead(lc_ctx_t *ctx, int overhead)
 {
+#ifdef HAVE_LIBLCRQ
 	return (ctx->rq_oh = overhead);
+#else
+	(void)ctx, (void)overhead;
+	return (errno = ENOTSUP), -1;
+#endif
 }
 
 int lc_channel_rq_overhead(lc_channel_t *chan, int overhead)
 {
+#ifdef HAVE_LIBLCRQ
 	return (chan->rq_oh = overhead);
-}
+#else
+	(void)chan, (void)overhead;
+	return (errno = ENOTSUP), -1;
 #endif
+}
 
 lc_socket_t *lc_channel_socket(lc_channel_t *chan)
 {
@@ -1098,7 +1116,7 @@ ssize_t lc_msg_send(lc_channel_t *chan, lc_message_t *msg)
 		return LC_ERROR_MALLOC;
 	}
 	memcpy(buf, head, sizeof(lc_message_head_t));
-	memcpy(buf + sizeof(lc_message_head_t), msg->data, len);
+	if (msg->data) memcpy(buf + sizeof(lc_message_head_t), msg->data, len);
 	len += sizeof(lc_message_head_t);
 
 	if (chan->coding & LC_CODE_SYMM) {
@@ -1155,8 +1173,8 @@ lc_channel_t *lc_socket_channel_by_address(lc_socket_t *sock, struct in6_addr *a
 	lc_ctx_t *lctx = lc_socket_ctx(sock);
 	lc_channel_t *p = NULL;
 	if (!pthread_mutex_lock(&lctx->mtx)) {
-		for (p = lctx->chan_list; p; p = p->next) {
-			if (p->sock != sock) continue;
+		for (p = aload(&lctx->chan_list); p; p = aload(&p->next)) {
+			if (aload(&p->sock) != sock) continue;
 			if (!memcmp(addr,& p->sa.sin6_addr, sizeof(struct in6_addr)))
 				break;
 		}
@@ -1209,6 +1227,83 @@ static ssize_t lc_channel_decoded(lc_channel_t *chan, unsigned char *m, unsigned
 }
 #endif
 
+#ifdef	HAVE_LINUX_NETLINK_H
+static void handle_netlink_msg(lc_socket_t *sock, struct nlmsghdr *msg)
+{
+	struct ifinfomsg *ifi = NLMSG_DATA(msg);
+	char ifname[IF_NAMESIZE];
+	lc_ctx_t *lctx = sock->ctx;
+	if (msg->nlmsg_type != RTM_NEWLINK) return;
+	/* new link detected, JOIN channels */
+	if_indextoname(ifi->ifi_index, ifname);
+	if (!pthread_mutex_lock(&lctx->mtx)) {
+		for (lc_channel_t *p = lctx->chan_list; p; p = p->next) {
+			if (p->sock != sock) continue;
+			lc_channel_join(p);
+		}
+		pthread_mutex_unlock(&lctx->mtx);
+	}
+}
+
+/* handle netlink event */
+static int recv_netlink_event(lc_socket_t *sock, int fd, struct sockaddr_nl *sa)
+{
+	struct nlmsghdr buf[8192/sizeof(struct nlmsghdr)];
+	struct iovec iov = { buf, sizeof(buf) };
+	struct msghdr msg = { sa, sizeof(*sa), &iov, 1, NULL, 0, 0 };
+	struct nlmsghdr *nh;
+	ssize_t len;
+	len = recvmsg(fd, &msg, 0);
+	if (len == -1) return -1;
+	for (nh = (struct nlmsghdr *)buf; NLMSG_OK(nh, len); nh = NLMSG_NEXT(nh, len)) {
+		if (nh->nlmsg_type == NLMSG_DONE) break; /* end of multipart message */
+		if (nh->nlmsg_type == NLMSG_ERROR) return -1;
+		else handle_netlink_msg(sock, nh);
+	}
+	return 0;
+}
+#endif
+
+static ssize_t socket_recvmsg(lc_socket_t *sock, struct msghdr *msg, int flags)
+{
+	enum {
+		FD_NETLNK,
+		FD_SOCKET
+	};
+	int fd = sock->sock;
+#ifdef	HAVE_LINUX_NETLINK_H
+	/* for IPv6 sockets which aren't bound, watch netlink */
+	if (!sock->ifx && sock->type == LC_SOCK_IN6) {
+		const nfds_t nfds = 2;
+		struct pollfd fds[nfds];
+		struct sockaddr_nl sa = {
+			.nl_family = AF_NETLINK,
+			.nl_groups = RTMGRP_LINK
+		};
+		int nlfd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
+		if (nlfd == -1) return -1;
+		if (bind(nlfd, (struct sockaddr *) &sa, sizeof(sa))) return -1;
+		fds[FD_NETLNK].fd = nlfd;
+		fds[FD_NETLNK].events = POLLIN;
+		fds[FD_SOCKET].fd = sock->sock;
+		fds[FD_SOCKET].events = POLLIN;
+		for (;;) {
+			fd = poll(fds, nfds, -1);
+			if (fd == -1) return -1;
+			if (fds[FD_NETLNK].revents) {
+				if (recv_netlink_event(sock, nlfd, &sa)) return -1;
+			}
+			if (fds[FD_SOCKET].revents) {
+				fd = fds[FD_SOCKET].fd;
+				break;
+			}
+		}
+		close(nlfd);
+	}
+#endif
+	return recvmsg(fd, msg, flags);
+}
+
 ssize_t lc_msg_recv(lc_socket_t *sock, lc_message_t *msg)
 {
 	ssize_t zi = 0;
@@ -1237,7 +1332,7 @@ recv_again:
 	if (setsockopt(sock->sock, IPPROTO_IPV6, IPV6_RECVPKTINFO, &opt, sizeof opt) == -1)
 		return -1;
 	pthread_testcancel();
-	if ((zi = recvmsg(sock->sock, &msgh, 0)) <= 0) return zi;
+	if ((zi = socket_recvmsg(sock, &msgh, 0)) <= 0) return zi;
 	for (cmsg = CMSG_FIRSTHDR(&msgh); cmsg; cmsg = CMSG_NXTHDR(&msgh, cmsg)) {
 		if (cmsg->cmsg_type == IPV6_PKTINFO) {
 			/* may not be aligned, copy */
@@ -1335,7 +1430,7 @@ static ssize_t lc_socket_recvmsg_if(lc_socket_t *sock, struct msghdr *msg, int f
 	}
 recv_again:
 	for (;;) {
-		bytes = recvmsg(sock->sock, msg, flags);
+		bytes = socket_recvmsg(sock, msg, flags);
 		if (bytes == -1) return -1;
 		if (sock->type != LC_SOCK_IN6) {
 			if (dst) {
@@ -1654,7 +1749,6 @@ static ssize_t socket_recvmsg_data(lc_socket_t *sock, struct msghdr *msg, size_t
 	lc_tlv_hash_t tlv_hash;
 	lc_channel_t *chan;
 	ssize_t rc;
-	ssize_t F = len;
 	int v = 0;
 	buf = malloc(buflen);
 	if (!buf) return -1;
@@ -1700,6 +1794,7 @@ recv_again:
 		if (chan->coding & LC_CODE_FEC_RQ) {
 #ifdef HAVE_LIBLCRQ
 #if HAVE_RQ_OTI
+			ssize_t F = len;
 			if (chan->coding & LC_CODE_SYMM) {
 				F += OH_SYMM, len += OH_SYMM;
 			}
@@ -1828,7 +1923,7 @@ ssize_t lc_socket_multi_recvmsg(lc_socket_t *sock, struct msghdr *msg, int flags
 		}
 		return socket_recvmsg_data(sock, msg, len, flags, dst);
 	}
-	return recvmsg(sock->sock, msg, flags);
+	return socket_recvmsg(sock, msg, flags);
 }
 
 ssize_t lc_socket_recvmsg(lc_socket_t *sock, struct msghdr *msg, int flags)
@@ -1848,11 +1943,10 @@ ssize_t lc_socket_recv(lc_socket_t *sock, void *buf, size_t len, int flags)
 	return lc_socket_multi_recv(sock, buf, len, flags, NULL);
 }
 
-#ifdef HAVE_LIBLCRQ
-#if HAVE_RQ_OTI
 /* peek at packet header to get OTI params */
 int lc_channel_oti_peek(lc_channel_t *chan, rq_oti_t *oti, rq_scheme_t *scheme)
 {
+#if (HAVE_LIBLCRQ && HAVE_RQ_OTI)
 	const size_t sz = ((chan->coding & LC_CODE_SYMM) ?
 		IPV6_MIN_MTU : sizeof *oti + sizeof *scheme) + HASHSIZE;
 	char buf[sz];
@@ -1865,9 +1959,11 @@ int lc_channel_oti_peek(lc_channel_t *chan, rq_oti_t *oti, rq_scheme_t *scheme)
 		chan->rq = rq_init(rq_oti_F(*oti), chan->T);
 	}
 	return rc;
+#else
+	(void)chan, (void)oti, (void) scheme;
+	return (errno = ENOTSUP), -1;
+#endif
 }
-#endif /* HAVE_RQ_OTI */
-#endif /* HAVE_LIBLCRQ */
 
 ssize_t lc_channel_recv(lc_channel_t *chan, void *buf, size_t len, int flags)
 {
@@ -2149,13 +2245,12 @@ int lc_channel_unbind(lc_channel_t *chan)
 
 static int lc_socket_bind_addr(lc_socket_t *sock, short port)
 {
-	int opt = 1;
+	const int opt = 1;
 	struct sockaddr_in6 any = {
 		.sin6_family = AF_INET6,
 		.sin6_addr = IN6ADDR_ANY_INIT,
 		.sin6_port = port
 	};
-
 	if ((setsockopt(sock->sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt))) == -1)
 		return LC_ERROR_SETSOCKOPT;
 
@@ -2163,12 +2258,10 @@ static int lc_socket_bind_addr(lc_socket_t *sock, short port)
 	if ((setsockopt(sock->sock, SOL_SOCKET, SO_REUSEPORT, &opt, sizeof(opt))) == -1)
 		return LC_ERROR_SETSOCKOPT;
 #endif
-
 	if (bind(sock->sock, (struct sockaddr *)&any, sizeof(struct sockaddr_in6)) == -1) {
-		/* ignore EINVAL "socket already bound" error */
-		if (errno != EINVAL) return LC_ERROR_SOCKET_BIND;
+		/* ignore EINVAL "socket already bound" and EADDRINUSE errors */
+		if (errno != EINVAL && errno != EADDRINUSE) return LC_ERROR_SOCKET_BIND;
 	}
-
 	return 0;
 }
 
@@ -2185,7 +2278,7 @@ int lc_channel_bind(lc_socket_t *sock, lc_channel_t *chan)
 	}
 
 	if (!rc) {
-		chan->sock = sock;
+		astor(&chan->sock, sock);
 		sock->bound++;
 		chan->sock->ccoding |= chan->coding;
 #ifdef HAVE_LIBLCRQ
