@@ -207,10 +207,10 @@ void mdex_hash_entry(unsigned char *hash, size_t hashlen, const char *path, stru
 		free(tmp);
 	}
 	if (sb) {
-		hash_update(&state, (unsigned char *)&sb->st_mode, sizeof(mode_t));
+		hash_update(&state, (unsigned char *)&sb->st_mode, sizeof(uint16_t));
 		hash_update(&state, (unsigned char *)&sb->st_uid, sizeof(uid_t));
 		hash_update(&state, (unsigned char *)&sb->st_gid, sizeof(gid_t));
-		hash_update(&state, (unsigned char *)&sb->st_mtim.tv_sec, sizeof(time_t));
+		hash_update(&state, (unsigned char *)&sb->st_mtim.tv_sec, sizeof(uint64_t));
 		long nsec = sb->st_mtim.tv_nsec >> 3 << 3; /* reduce precision */
 		hash_update(&state, (unsigned char *)&nsec, sizeof(long));
 	}
@@ -237,10 +237,10 @@ void mdex_tree_hash_sb(unsigned char *hash, size_t hashlen, mtree_t *mtree, size
 		hash_update(&state, (unsigned char *)&n, sizeof(n));
 	}
 	if (sb) {
-		hash_update(&state, (unsigned char *)&sb->st_mode, sizeof(mode_t));
+		hash_update(&state, (unsigned char *)&sb->st_mode, sizeof(uint16_t));
 		hash_update(&state, (unsigned char *)&sb->st_uid, sizeof(uid_t));
 		hash_update(&state, (unsigned char *)&sb->st_gid, sizeof(gid_t));
-		hash_update(&state, (unsigned char *)&sb->st_mtim.tv_sec, sizeof(time_t));
+		hash_update(&state, (unsigned char *)&sb->st_mtim.tv_sec, sizeof(uint64_t));
 		long nsec = sb->st_mtim.tv_nsec >> 3 << 3; /* reduce precision */
 		hash_update(&state, (unsigned char *)&nsec, sizeof(long));
 	}
@@ -293,8 +293,10 @@ static int mdex_tree(mdex_t *mdex, mtree_t *mtree, const char *path, struct stat
 		memcpy(data->tree, sharepath, namesz);
 	}
 	if (sb) {
-		data->atime = sb->st_atim;
-		data->mtime = sb->st_mtim;
+		data->atime_s = sb->st_atim.tv_sec;
+		data->atime_n = sb->st_atim.tv_nsec;
+		data->mtime_s = sb->st_mtim.tv_sec;
+		data->mtime_n = sb->st_mtim.tv_nsec;
 		data->mode = sb->st_mode;
 		data->uid = sb->st_uid;
 		data->gid = sb->st_gid;
@@ -711,6 +713,7 @@ int mdex_del(mdex_t *mdex, unsigned char *hash, size_t hashlen)
 mdex_t *mdex_init(size_t entries)
 {
 	mdex_t *mdex = calloc(1, sizeof (mdex_t));
+	mdex->readers = 1;
 	if (mdex) {
 		mdex->pagesz = (size_t)sysconf(_SC_PAGESIZE); /* cannot be < 1 */
 		mdex->entries = entries;
@@ -720,7 +723,14 @@ mdex_t *mdex_init(size_t entries)
 	return mdex;
 }
 
-void mdex_free(mdex_t *mdex)
+mdex_t *mdex_acquire(mdex_t *mdex)
+{
+	if (!mdex || aload(&mdex->deleted)) return NULL;
+	fetch_add(&mdex->readers, 1); /* increase reference count */
+	return mdex;
+}
+
+void mdex_free_all(mdex_t *mdex)
 {
 	void *ptr;
 	while ((ptr = mdex_stack_pop(&mdex->alloc_stack))) free(ptr);
@@ -731,3 +741,21 @@ void mdex_free(mdex_t *mdex)
 	free(mdex->rootdir);
 	free(mdex);
 }
+
+void mdex_release(mdex_t *mdex)
+{
+	/* decrease reference count, fetch updated value */
+	int refs = add_fetch(&mdex->readers, -1);
+	int deleted = aload(&mdex->deleted);
+	if (!refs && deleted == DELETE_LATER) {
+		/* no readers left, try to delete */
+		if (CAE(&mdex->deleted, &mdex, DELETE_IN_PROGRESS)) {
+			mdex_free_all(mdex);
+		}
+	}
+}
+
+void mdex_free(mdex_t *mdex)
+{
+	mdex_release(mdex);
+}
