@@ -1,5 +1,5 @@
 /* SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only */
-/* Copyright (c) 2023-2024 Brett Sheffield <bacs@librecast.net> */
+/* Copyright (c) 2023-2025 Brett Sheffield <bacs@librecast.net> */
 
 #define _GNU_SOURCE
 #include <librecast_pvt.h>
@@ -18,6 +18,10 @@
 #include <utime.h>
 #endif
 
+#define DEBUG_SYNC(...) if (lctx->debug & LCTX_DEBUG_SYNCFILE) fprintf(lctx->stream, __VA_ARGS__)
+
+static sem_t semjob;
+
 struct syncjob {
 	sem_t done;
 	lc_ctx_t *lctx;
@@ -157,8 +161,10 @@ static ssize_t buildtree(ssize_t rc, lc_ctx_t *lctx, mtree_t *tree, net_tree_t *
 		len -= sizeof(struct net_tree_s) + data->namesz;
 		if (parts > 1) {
 			struct stat sb = {
-				.st_atim = data->atime,
-				.st_mtim = data->mtime,
+				.st_atim.tv_sec = data->atime_s,
+				.st_atim.tv_nsec = data->atime_n,
+				.st_mtim.tv_sec = data->mtime_s,
+				.st_mtim.tv_nsec = data->mtime_n,
 				.st_mode = data->mode,
 				.st_uid = data->uid,
 				.st_gid = data->gid,
@@ -275,7 +281,7 @@ ssize_t lc_recvchunk(lc_ctx_t *lctx, unsigned char *hash, void *data, const size
 	if (lc_channel_join(chan) == -1) { rc = -1; goto err_channel_free; }
 	pthread_cleanup_push((void (*)(void *))_lc_channel_part, chan);
 	rc = lc_channel_recv(chan, data, len, flags);
-	if (stats && rc > 0) stats->byt_gross += rc;
+	if (stats && rc > 0) add_fetch(&stats->byt_gross, rc);
 	pthread_cleanup_pop(1); /* lc_channel_part */
 err_channel_free:
 	pthread_cleanup_pop(1); /* lc_channel_free */
@@ -367,6 +373,7 @@ static char *concat_mdex_root_file(const char *root, const char *file)
 
 static ssize_t share_send_file(lc_share_t *share, struct in6_addr *grp, mdex_entry_t *entry, int coding)
 {
+	lc_ctx_t *lctx = share->lctx;
 	const struct mdex_file_s *file = &entry->file;
 	struct mdex_mem_s mem = { .size = file->size };
 	struct stat sbs = {0};
@@ -378,8 +385,7 @@ static ssize_t share_send_file(lc_share_t *share, struct in6_addr *grp, mdex_ent
 	if ((smap = lc_mmapfile(path, &sz, PROT_READ, MAP_PRIVATE, 0, &sbs)) == MAP_FAILED)
 		goto err_free_path;
 	mem.data = smap + file->off;
-	if (share->lctx->debug & LCTX_DEBUG_SYNCFILE)
-		fprintf(share->lctx->stream, "%s\n", file->file.name);
+	DEBUG_SYNC("%s\n", file->file.name);
 	rc = share_send_data(share, grp, &mem, coding);
 	munmap(smap, sz);
 err_free_path:
@@ -389,6 +395,7 @@ err_free_path:
 
 static ssize_t share_send_dir(lc_share_t *share, struct in6_addr *grp, mdex_entry_t *entry, int coding)
 {
+	lc_ctx_t *lctx = share->lctx;
 	const struct mdex_file_s *file = &entry->file;
 	net_tree_t *dir;
 	struct mdex_mem_s mem = {0};
@@ -403,16 +410,17 @@ static ssize_t share_send_dir(lc_share_t *share, struct in6_addr *grp, mdex_entr
 	tree = dir->tree + namesz;
 	if (file->size) memcpy(tree, file->file.dir, sizeof(hash_t) * file->size);
 	dir->size = htobe64(sz);
-	dir->mtime = file->file.sb->st_mtim;
-	dir->atime = file->file.sb->st_atim;
+	dir->mtime_s = file->file.sb->st_mtim.tv_sec;
+	dir->mtime_n = file->file.sb->st_mtim.tv_nsec;
+	dir->atime_s = file->file.sb->st_atim.tv_sec;
+	dir->atime_n = file->file.sb->st_atim.tv_nsec;
 	dir->mode = file->file.sb->st_mode;
 	dir->uid = file->file.sb->st_uid;
 	dir->gid = file->file.sb->st_gid;
 	dir->namesz = htons(namesz);
 	mem.data = dir;
 	mem.size = sz;
-	if (share->lctx->debug & LCTX_DEBUG_SYNCFILE)
-		fprintf(share->lctx->stream, "%s\n", file->file.name);
+	DEBUG_SYNC("%s\n", file->file.name);
 	rc = share_send_data(share, grp, &mem, coding);
 	free(dir);
 	return rc;
@@ -422,6 +430,7 @@ static ssize_t share_send_dir(lc_share_t *share, struct in6_addr *grp, mdex_entr
  * header metadata */
 static ssize_t share_send_link(lc_share_t *share, struct in6_addr *grp, mdex_entry_t *entry, int coding)
 {
+	lc_ctx_t *lctx = share->lctx;
 	const struct mdex_file_s *file = &entry->file;
 	net_tree_t *dir;
 	char *buf, *linkpath = NULL;
@@ -450,16 +459,17 @@ static ssize_t share_send_link(lc_share_t *share, struct in6_addr *grp, mdex_ent
 	tree = dir->tree + namesz;
 	memcpy(tree, buf, (size_t)rc);
 	dir->size = htobe64(sz);
-	dir->mtime = file->file.sb->st_mtim;
-	dir->atime = file->file.sb->st_atim;
+	dir->mtime_s = file->file.sb->st_mtim.tv_sec;
+	dir->mtime_n = file->file.sb->st_mtim.tv_nsec;
+	dir->atime_s = file->file.sb->st_atim.tv_sec;
+	dir->atime_n = file->file.sb->st_atim.tv_nsec;
 	dir->mode = file->file.sb->st_mode;
 	dir->uid = file->file.sb->st_uid;
 	dir->gid = file->file.sb->st_gid;
 	dir->namesz = htons(namesz);
 	mem.data = dir;
 	mem.size = sz;
-	if (share->lctx->debug & LCTX_DEBUG_SYNCFILE)
-		fprintf(share->lctx->stream, "%s\n", file->file.name);
+	DEBUG_SYNC("%s\n", file->file.name);
 	rc = share_send_data(share, grp, &mem, coding);
 	free(dir);
 err_free_linkpath:
@@ -471,7 +481,8 @@ err_free_buf:
 
 static void *share_send_thread(void *arg)
 {
-	lc_share_t *share = (lc_share_t *)arg;
+	lc_share_t *share = lc_share_acquire((lc_share_t *)arg);
+	if (!share) return NULL;
 	q_job_t job = {0};
 	while (1) {
 		pthread_testcancel();
@@ -538,6 +549,7 @@ lc_share_t * lc_share(lc_ctx_t *lctx, mdex_t *mdex, unsigned int ifx, lc_stat_t
 	/* allocate & initialize share */
 	share = calloc(1, sizeof (lc_share_t));
 	if (!share) return NULL;
+	share->readers = 1;
 	share->tsend = calloc(nthread_send, sizeof(pthread_t));
 	if (!share->tsend) { err = errno; goto err_free_share; }
 	share->lctx = lctx;
@@ -575,19 +587,42 @@ err_free_share:
 #endif /* HAVE_MLD */
 }
 
+lc_share_t *lc_share_acquire(lc_share_t *share)
+{
+	if (!share || aload(&share->deleted)) return NULL;
+	fetch_add(&share->readers, 1); /* increase reference count */
+	return share;
+}
+
+void lc_share_release(lc_share_t *share)
+{
+	/* decrease reference count, fetch updated value */
+	int refs = add_fetch(&share->readers, -1);
+	int deleted = aload(&share->deleted);
+	if (!refs && deleted == DELETE_LATER) {
+		/* no readers left, try to delete */
+		if (CAE(&share->deleted, &share, DELETE_IN_PROGRESS)) {
+			free(share);
+		}
+	}
+}
+
 void lc_unshare(lc_share_t *share)
 {
 #ifdef HAVE_MLD
 	int err = errno;
 	q_job_t job = {0};
 	q_pool_destroy(share->tsend, share->nthread_send);
+	while (share->nthread_send--) lc_share_release(share);
 	free(share->tsend);
 	while (!q_trywait(&share->q, &job)) free(job.arg);
 	(void)q_free(&share->q);
 	mld_stop(share->mld);
 	mld_watch_del(share->watch);
 	mld_free(share->mld);
-	free(share);
+	int deleted = 0;
+	if (CAE(&share->deleted, &deleted, DELETE_LATER))
+		lc_share_release(share);
 	errno = err;
 #else
 	(void)share;
@@ -703,16 +738,24 @@ static int setfilestat(const char *pathname, struct stat *sb, int flags)
 #ifdef HAVE_RQ_OTI
 static int setfilemeta(const char *pathname, net_tree_t *data, int flags)
 {
-	struct timespec *atime, *mtime;
+	struct timespec atime = {0};
+	struct timespec mtime = {0};
 	int uid, gid;
 	int rc = 0;
-	atime = (flags & SYNC_ATIME) ? &data->atime : NULL;
-	mtime = (flags & SYNC_MTIME) ? &data->mtime : NULL;
+	if (flags & SYNC_ATIME) {
+		atime.tv_sec = data->atime_s;
+		atime.tv_nsec = data->atime_n;
+	}
+	if (flags & SYNC_MTIME) {
+		mtime.tv_sec = data->mtime_s;
+		mtime.tv_nsec = data->mtime_n;
+	}
 	uid = (flags & SYNC_OWNER) ? (int)data->uid : -1;
 	gid = (flags & SYNC_GROUP) ? (int)data->gid : -1;
 	if ((uid | gid) != -2) rc += chown(pathname, uid, gid);
 	if (flags & SYNC_MODE) rc += chmod(pathname, data->mode & 0777);
-	if (atime || mtime) rc += utimeset(pathname, atime, mtime);
+	if ((flags & SYNC_ATIME) || (flags & SYNC_MTIME))
+		rc += utimeset(pathname, &atime, &mtime);
 	return rc;
 }
 #endif
@@ -748,8 +791,10 @@ static int copydirlocal(mdex_t *sdex, mdex_t *ddex, mdex_hash_t sdir, size_t top
 		rc = mtree_init(&mtree, data->size);
 		if (rc == -1) goto err_free_sharepath;
 		memcpy(mtree.tree, data->tree + data->namesz, (size_t)mtree.nodes * (size_t)HASHSIZE);
-		sb.st_atim = data->atime;
-		sb.st_mtim = data->mtime;
+		sb.st_atim.tv_sec = data->atime_s;
+		sb.st_atim.tv_nsec = data->atime_n;
+		sb.st_mtim.tv_sec = data->mtime_s;
+		sb.st_mtim.tv_nsec = data->mtime_n;
 		sb.st_mode = data->mode;
 		sb.st_uid = data->uid;
 		sb.st_gid = data->gid;
@@ -799,6 +844,7 @@ static int syncdirlocal(mdex_t *sdex, mdex_t *ddex, const char *src, const char
 		if (rc == -1) return -1;
 	}
 	else ddir = dentry.file.file.dir;
+	if (sdir == NULL && ddir == NULL) return 0;
 	if (sentry.file.size == dentry.file.size) {
 		size_t sz = sentry.file.size * sizeof(*sdir);
 		if (!memcmp(sdir, ddir, sz)) return 0; /* directories match */
@@ -977,6 +1023,7 @@ static void * sync_chunk(void *arg)
 	}
 	while (memcmp(hash, job->hash, HASHSIZE));
 	sem_post(&job->done);
+	sem_post(&semjob);
 	return NULL;
 }
 
@@ -992,8 +1039,10 @@ inline static ssize_t syncfile(lc_ctx_t *lctx, unsigned char *map, mtree_t *stre
 	job = calloc(diffs, sizeof(struct syncjob));
 	if (!job) return -1;
 	/* queue jobs to sync each differing chunk */
+	sem_init(&semjob, 0, SYNC_MAX_RECVERS);
 	for (size_t z = 0; z < stree->chunks; z++) {
 		if (isset(map, z)) {
+			sem_wait(&semjob);
 			size_t off = z * MTREE_CHUNKSIZE;
 			job[j].lctx = lctx;
 			job[j].sz = MIN(MTREE_CHUNKSIZE, len - off);
@@ -1019,6 +1068,7 @@ inline static ssize_t syncfile(lc_ctx_t *lctx, unsigned char *map, mtree_t *stre
 			err = job->err;
 		}
 	}
+	sem_destroy(&semjob);
 	if (rc != -1) rc = byt;
 	else err = errno;
 	free(job);
@@ -1046,7 +1096,7 @@ static char *mksubdir(lc_ctx_t *lctx, const char *pathname, net_tree_t *data, si
 	if (!subdir) return NULL;
 	rc = mkdir(subdir, 0755); /* create writable, will set mode later */
 	if (rc == -1) goto err_free_subdir;
-	if (lctx->debug & LCTX_DEBUG_SYNCFILE) fprintf(lctx->stream, "%s\n", subdir);
+	DEBUG_SYNC("%s\n", subdir);
 	return subdir;
 err_free_subdir:
 	rc = errno;
@@ -1080,7 +1130,7 @@ static ssize_t synclink(lc_ctx_t *lctx, const char *pathname, net_tree_t *data,
 	if (rc == -1) {
 		if (errno == EEXIST) rc = 0;
 	}
-	else if (lctx->debug & LCTX_DEBUG_SYNCFILE) fprintf(lctx->stream, "%s\n", linkpath);
+	else DEBUG_SYNC("%s\n", linkpath);
 	/* XXX: do not set mode for symlinks */
 	if (!rc) setfilemeta(linkpath, data, (flags | SYNC_MODE) ^ SYNC_MODE);
 	free(target);
@@ -1125,10 +1175,10 @@ static mdex_t *index_dest(lc_ctx_t *lctx, const char *pathname, net_tree_t *data
 	ddex = mdex_init(0);
 	if (!ddex) return NULL;
 	subdir = mkentryname(pathname, data, trim);
-	if (lctx->debug & LCTX_DEBUG_SYNCFILE) fprintf(lctx->stream, "hashing directory %s... ", subdir);
+	DEBUG_SYNC("hashing directory %s... ", subdir);
 	rc = mdex_addfile(ddex, subdir, q, MDEX_RECURSE);
 	if (rc == -1) goto err_free_subdir;
-	if (lctx->debug & LCTX_DEBUG_SYNCFILE) fprintf(lctx->stream, "done.\n");
+	DEBUG_SYNC("done.\n");
 	free(subdir);
 	return ddex;
 err_free_subdir:
@@ -1210,6 +1260,7 @@ ssize_t lc_syncfile(lc_ctx_t *lctx, unsigned char *hash, const char *pathname,
 			goto err_free_q;
 		}
 	}
+	DEBUG_SYNC("%s\n", mapfile);
 	if (!len) {
 		/* zero-length file */
 		rc = creat(mapfile, 0755);
@@ -1218,7 +1269,6 @@ ssize_t lc_syncfile(lc_ctx_t *lctx, unsigned char *hash, const char *pathname,
 			goto err_free_q;
 		}
 		close(rc);
-		if (lctx->debug & LCTX_DEBUG_SYNCFILE) fprintf(lctx->stream, "%s\n", mapfile);
 		goto do_setfilemeta;
 	}
 	dst = lc_mmapfile(mapfile, &len, PROT_READ|PROT_WRITE, MAP_SHARED, 0, &sbd);
@@ -1239,7 +1289,6 @@ ssize_t lc_syncfile(lc_ctx_t *lctx, unsigned char *hash, const char *pathname,
 	}
 	map = mtree_diff_subtree(&stree, &dtree, 0, 1);
 	if (map) {
-		if (lctx->debug & LCTX_DEBUG_SYNCFILE) fprintf(lctx->stream, "%s\n", fname);
 		rc = syncfile(lctx, map, &stree, dst, len, q, stats, opt, 0);
 		if (rc != -1) {
 			rc = byt;
