@@ -119,7 +119,16 @@ void test_log(char *msg, ...)
 	b = malloc(_vscprintf(msg, argp) + 1);
 	pthread_cleanup_push(free, b);
 	vsprintf(b, msg, argp);
-	sem_wait(&log_lock);
+
+	/* In order to keep the test logs clean we try to wait for other threads
+	 * to finish before logging, but if we need to wait more than a second,
+	 * give up and log anyway. A deadlock between two test threads is best
+	 * avoided. */
+	struct timespec ts;
+	clock_gettime(CLOCK_REALTIME, &ts);
+	ts.tv_sec++;
+	sem_timedwait(&log_lock, &ts);
+
 	fprintf(stderr, "%s", b);
 	fflush(stderr);
 	sem_post(&log_lock);
